@@ -63,7 +63,7 @@ int bc_finalizer::run() {
 
 	// workaround for some problems on r6xx/7xx
 	// add ALU NOP to each vertex shader
-	if (!ctx.is_egcm() && sh.target == TARGET_VS) {
+	if (!ctx.is_egcm() && (sh.target == TARGET_VS || sh.target == TARGET_ES)) {
 		cf_node *c = sh.create_clause(NST_ALU_CLAUSE);
 
 		alu_group_node *g = sh.create_alu_group();
@@ -83,14 +83,18 @@ int bc_finalizer::run() {
 		last_cf = c;
 	}
 
-	if (last_cf->bc.op_ptr->flags & CF_ALU) {
+	if (!ctx.is_cayman() && last_cf->bc.op_ptr->flags & CF_ALU) {
 		last_cf = sh.create_cf(CF_OP_NOP);
 		sh.root->push_back(last_cf);
 	}
 
-	if (ctx.is_cayman())
-		last_cf->insert_after(sh.create_cf(CF_OP_CF_END));
-	else
+	if (ctx.is_cayman()) {
+		if (!last_cf) {
+			cf_node *c = sh.create_cf(CF_OP_CF_END);
+			sh.root->push_back(c);
+		} else
+			last_cf->insert_after(sh.create_cf(CF_OP_CF_END));
+	} else
 		last_cf->bc.end_of_program = 1;
 
 	for (unsigned t = EXP_PIXEL; t < EXP_TYPE_COUNT; ++t) {
@@ -394,6 +398,64 @@ void bc_finalizer::finalize_alu_src(alu_group_node* g, alu_node* a) {
 	}
 }
 
+void bc_finalizer::copy_fetch_src(fetch_node &dst, fetch_node &src, unsigned arg_start)
+{
+	int reg = -1;
+
+	for (unsigned chan = 0; chan < 4; ++chan) {
+
+		dst.bc.dst_sel[chan] = SEL_MASK;
+
+		unsigned sel = SEL_MASK;
+
+		value *v = src.src[arg_start + chan];
+
+		if (!v || v->is_undef()) {
+			sel = SEL_MASK;
+		} else if (v->is_const()) {
+			literal l = v->literal_value;
+			if (l == literal(0))
+				sel = SEL_0;
+			else if (l == literal(1.0f))
+				sel = SEL_1;
+			else {
+				sblog << "invalid fetch constant operand  " << chan << " ";
+				dump::dump_op(&src);
+				sblog << "\n";
+				abort();
+			}
+
+		} else if (v->is_any_gpr()) {
+			unsigned vreg = v->gpr.sel();
+			unsigned vchan = v->gpr.chan();
+
+			if (reg == -1)
+				reg = vreg;
+			else if ((unsigned)reg != vreg) {
+				sblog << "invalid fetch source operand  " << chan << " ";
+				dump::dump_op(&src);
+				sblog << "\n";
+				abort();
+			}
+
+			sel = vchan;
+
+		} else {
+			sblog << "invalid fetch source operand  " << chan << " ";
+			dump::dump_op(&src);
+			sblog << "\n";
+			abort();
+		}
+
+		dst.bc.src_sel[chan] = sel;
+	}
+
+	if (reg >= 0)
+		update_ngpr(reg);
+
+	dst.bc.src_gpr = reg >= 0 ? reg : 0;
+}
+
 void bc_finalizer::emit_set_grad(fetch_node* f) {
 
 	assert(f->src.size() == 12);
@@ -405,70 +467,27 @@ void bc_finalizer::emit_set_grad(fetch_node* f) {
 		fetch_node *n = sh.create_fetch();
 		n->bc.set_op(ops[op]);
 
-		// FIXME extract this loop into a separate method and reuse it
-
-		int reg = -1;
-
 		arg_start += 4;
 
-		for (unsigned chan = 0; chan < 4; ++chan) {
-
-			n->bc.dst_sel[chan] = SEL_MASK;
-
-			unsigned sel = SEL_MASK;
-
-			value *v = f->src[arg_start + chan];
-
-			if (!v || v->is_undef()) {
-				sel = SEL_MASK;
-			} else if (v->is_const()) {
-				literal l = v->literal_value;
-				if (l == literal(0))
-					sel = SEL_0;
-				else if (l == literal(1.0f))
-					sel = SEL_1;
-				else {
-					sblog << "invalid fetch constant operand  " << chan << " ";
-					dump::dump_op(f);
-					sblog << "\n";
-					abort();
-				}
-
-			} else if (v->is_any_gpr()) {
-				unsigned vreg = v->gpr.sel();
-				unsigned vchan = v->gpr.chan();
-
-				if (reg == -1)
-					reg = vreg;
-				else if ((unsigned)reg != vreg) {
-					sblog << "invalid fetch source operand  " << chan << " ";
-					dump::dump_op(f);
-					sblog << "\n";
-					abort();
-				}
-
-				sel = vchan;
-
-			} else {
-				sblog << "invalid fetch source operand  " << chan << " ";
-				dump::dump_op(f);
-				sblog << "\n";
-				abort();
-			}
-
-			n->bc.src_sel[chan] = sel;
-		}
-
-		if (reg >= 0)
-			update_ngpr(reg);
-
-		n->bc.src_gpr = reg >= 0 ? reg : 0;
+		copy_fetch_src(*n, *f, arg_start);
 
 		f->insert_before(n);
 	}
 
 }
 
+void bc_finalizer::emit_set_texture_offsets(fetch_node &f) {
+	assert(f.src.size() == 8);
+
+	fetch_node *n = sh.create_fetch();
+
+	n->bc.set_op(FETCH_OP_SET_TEXTURE_OFFSETS);
+
+	copy_fetch_src(*n, f, 4);
+
+	f.insert_before(n);
+}
+
 void bc_finalizer::finalize_fetch(fetch_node* f) {
 
 	int reg = -1;
@@ -483,6 +502,8 @@ void bc_finalizer::finalize_fetch(fetch_node* f) {
 		src_count = 1;
 	} else if (flags & FF_USEGRAD) {
 		emit_set_grad(f);
+	} else if (flags & FF_USE_TEXTURE_OFFSETS) {
+		emit_set_texture_offsets(*f);
 	}
 
 	for (unsigned chan = 0; chan < src_count; ++chan) {
@@ -695,7 +716,7 @@ void bc_finalizer::finalize_cf(cf_node* c) {
 		c->bc.rw_gpr = reg >= 0 ? reg : 0;
 		c->bc.comp_mask = mask;
 
-		if ((flags & CF_RAT) && (c->bc.type & 1)) {
+		if (((flags & CF_RAT) || (!(flags & CF_STRM))) && (c->bc.type & 1)) {
 
 			reg = -1;
 
@@ -770,7 +791,6 @@ void bc_finalizer::update_ngpr(unsigned gpr) {
 unsigned bc_finalizer::get_stack_depth(node *n, unsigned &loops,
                                            unsigned &ifs, unsigned add) {
 	unsigned stack_elements = add;
-	bool has_non_wqm_push_with_loops_on_stack = false;
 	bool has_non_wqm_push = (add != 0);
 	region_node *r = n->is_region() ?
 			static_cast<region_node*>(n) : n->get_parent_region();
@@ -781,8 +801,6 @@ unsigned bc_finalizer::get_stack_depth(node *n, unsigned &loops,
 	while (r) {
 		if (r->is_loop()) {
 			++loops;
-			if (has_non_wqm_push)
-				has_non_wqm_push_with_loops_on_stack = true;
 		} else {
 			++ifs;
 			has_non_wqm_push = true;
@@ -795,17 +813,30 @@ unsigned bc_finalizer::get_stack_depth(node *n, unsigned &loops,
 	switch (ctx.hw_class) {
 	case HW_CLASS_R600:
 	case HW_CLASS_R700:
+		// If any non-WQM push is invoked, 2 elements should be reserved.
 		if (has_non_wqm_push)
 			stack_elements += 2;
 		break;
 	case HW_CLASS_CAYMAN:
+		// If any stack operation is invoked, 2 elements should be reserved
 		if (stack_elements)
 			stack_elements += 2;
 		break;
 	case HW_CLASS_EVERGREEN:
-		if (has_non_wqm_push_with_loops_on_stack)
+		// According to the docs we need to reserve 1 element for each of the
+		// following cases:
+		//   1) non-WQM push is used with WQM/LOOP frames on stack
+		//   2) ALU_ELSE_AFTER is used at the point of max stack usage
+		// NOTE:
+		// It was found that the conditions above are not sufficient, there are
+		// other cases where we also need to reserve stack space, that's why
+		// we always reserve 1 stack element if we have non-WQM push on stack.
+		// Condition 2 is ignored for now because we don't use this instruction.
+		if (has_non_wqm_push)
 			++stack_elements;
 		break;
+	case HW_CLASS_UNKNOWN:
+		assert(0);
 	}
 	return stack_elements;
 }
